#coding: utf-8

import os, sys, json
import requests, traceback
import argparse, shlex, platform
import importlib

from prompt_toolkit import prompt
from prompt_toolkit.output.win32 import NoConsoleScreenBufferError
from requests import get, post, put, delete

if __name__ == "__main__":
    from utils import gitee
    from utils.backend import SqleeColumn, SqleeTable
    from utils.exceptions import RepositoryNotFoundError, TableNotFoundError, RepositoryNotFound
    from utils.urlparse import URL
    from utils.crypto import CryptoHandler
    from config import __version__
else:
    from .utils import gitee
    from .utils.backend import SqleeColumn, SqleeTable
    from .utils.exceptions import RepositoryNotFoundError, TableNotFoundError, RepositoryNotFound
    from .utils.urlparse import URL
    from .utils.crypto import CryptoHandler
    from .config import __version__

class termios:
    class error(RuntimeError):
        pass
Windows = True if platform.system() == "Windows" else False

class objects:
    def __init__(self, obj=None):
        self.obj = obj

    def all(self):
        return self.obj.all_tables()
    
    def get(self, name=None, directly_load=True):
        if name in self.obj.repo.list_folder(path=""):
            return SqleeTable(name=name, repo=self.obj.repo, directly_load=directly_load)
        raise TableNotFoundError(name)

    def create(self, name=None, namespace=None):
        return self.obj.create_table(name=name, namespace=namespace)

    def delete(self, *args, **kwargs):
        return self.obj.drop_table(*args, **kwargs)
    
    def count(self):
        return len(self.obj.repo.list_folder(path="", detail=True))

    @property
    def count(self):
        return len(self.obj.repo.list_folder(path="", detail=True))

class SqleeRepository:
    access_token = None
    repo = None
    owner = None
    def __init__(self, access_token, repo, owner, auto_make=False):
        self.access_token = access_token
        if auto_make:
            print(RepositoryNotFound(repo))
            gitee.make_repo(name=repo, user=owner, token=access_token)
        self.repo = gitee.GiteeRepo(token=access_token, user=owner, repo=repo)
        self.owner = owner
        self.objects = objects(obj=self)

    def all_tables(self):
        datas = self.repo.list_folder(path="", detail=True)
        tables = [SqleeTable(name=data['name'], repo=self.repo) for data in datas if data['type'] == 'dir']
        return tables

    def create_table(self, name=None, namespace=None):
        if not namespace:
            raise ValueError("未提供足够的参数.")
        for value in ["data", "length", "update", "delete",
                      "namespace", "count", "sync", "repo"]:
            if value in namespace:
                raise ValueError("命名域不可以被命名为: %s." % value)
        make = self.repo.make_folder(path=name)
        self.repo.upload_file(path=URL(name)/".namespace", content=json.dumps(namespace))
        return make

    def drop_table(self, name=None):
        return self.repo.drop_folder(path=name)

    def clear(self, sure=False):
        if not sure:
            confirm = input("你确定删除本数据库(同时删除云端数据)吗? (Y/N) ")
            if confirm == "Y" or confirm == "y":
                pass
            else:
                print("忽略.")
                return
        url = f"https://gitee.com/api/v5/repos/{self.owner}/{self.repo}"

        requests.delete(
            url = url,
            data = {"access_token": self.access_token}
        )
        del self
        return True
    
    def interact(self):
        def hotreload():
            import os, sys
            os.system("python %s" % sys.argv[0])

        print("SQLEE %s\nCopyright © 中国左旋联盟 × Freet安全组" % (__version__))
        all_tables = self.all_tables
        get_all_tables = self.all_tables
        tables = self.all_tables
        create_table = self.create_table
        create = self.create_table
        while True:
            try:
                cmd = prompt('SQLEE>>> ')
                if cmd.replace(" ", "") == "exit()":
                    return
            except NoConsoleScreenBufferError:
                if Windows:
                    print("SQLEE CONSOLE必须在 'cmd.exe' 中运行 !")
                    restart = input("是否要在CMD中重新执行本程序(或脚本)? (Y/N) ")
                    restart = True if restart == "Y" else False
                    if restart:
                        os.system("python %s" % sys.argv[0])
                else:
                    print("你的终端暂不支持运行HFCONSOLE.")
                exit()
            except termios.error:
                print('[-] Termios 错误, 你的终端不支持SQLEE CONSOLE或由于SQLEE CONSOLE开启时终端显示大小被调整, 请检查终端设置并重试.')
                exit()
            except KeyboardInterrupt:
                print("执行 'exit()' 来退出该会话.")
                continue
            except EOFError:
                print("不要使用 'Ctrl+D'(EOF) 来退出, 用 'exit()' 来代替它.")
                continue

            try:
                exec(cmd)
            except:
                traceback.print_exc()

if __name__ == '__main__':
    """t_s = Sqlee(
        access_token="f8eca055b26c3e4c64176d4c9f66902b",
        owner="freetbash",
        repo="sqleedb",
    )"""
    db = SqleeRepository(
        access_token = "1895956f770eb0e4d08013ee4b753203",
        owner = "fu050409",
        repo = "TEST_API",
    )



#coding: utf-8
__author__ = "Entropy <fu050409@163.com>"

class URL:
    def __init__(self, url="https://gitee.com/"):
        self.url = url if url != None else self.url
        self.url = self.url if self.url[-1] == "/" else self.url + "/"
        
    def __add__(self, value):
        value = str(value)
        value = value if value[-1] == "/" else value + "/"
        return URL(url=self.url + value)
        
    def __truediv__(self, value):
        value = str(value)
        value = value if value[-1] == "/" else value + "/"
        return URL(url=self.url + value)

    def __str__(self):
        return self.url



from .sqlee import SqleeRepository
from .config import __version__
from .utils.exceptions import VersionError

__name__ = "sqlee"
__version__ = __version__
__all__ = [
    'connect', 'utils',
    ]
__author__ = "Entropy <fu050409@163.com>"
__incantation__ = "嗡嘛呢叭咪吽"

def connect(name=None, user=None, token=None, *args, **kwargs):
    if "owner" in kwargs or "repo" in kwargs:
        raise VersionError("该连接数据库方式已在2.0.0版本后被弃用.")
    return SqleeRepository(access_token=token, repo=name, owner=user, *args, **kwargs)



#coding: utf-8
from sqlee.utils.django import models
from django.forms.models import model_to_dict
from threading import Thread

import argparse
import os
import sqlee
import sys
import json
import requests

if __name__ == "__main__":
    from verify import verify_django
else:
    from .verify import verify_django

def init():
    url = json.loads(requests.get("https://gitee.com/api/v5/repos/{user}/{name}?access_token={token}".format(
        user = settings.SQLEE_SETTINGS["OWNER"],
        name = settings.SQLEE_SETTINGS["NAME"],
        token = settings.SQLEE_SETTINGS["TOKEN"]
        )).text)
    if "message" in url:
        if url["message"] == "Not Found Project":
            print("开始创建数据库...")
            sqlee.utils.gitee.make_repo(token=settings.SQLEE_SETTINGS["TOKEN"], user=settings.SQLEE_SETTINGS["OWNER"], name=settings.SQLEE_SETTINGS["NAME"]).text
        else:
            raise Exception(url["message"])
    else:
        return True
            
def migrate():
    repo = sqlee.connect(token=settings.SQLEE_SETTINGS["TOKEN"], repo=settings.SQLEE_SETTINGS["NAME"], owner=settings.SQLEE_SETTINGS["OWNER"])
    sys.path.append(settings.BASE_DIR)

    for app in settings.INSTALLED_APPS:
        try:
            print("正在解析模块'%s'内的SQLEE数据库声明文件." % (app))
            sqlee_models = __import__("%s.sqlee_models" % (app)).sqlee_models
        except ModuleNotFoundError:
            print("模块'%s'中未找到SQLEE数据库声明文件." % (app))
            continue
        for model in dir(sqlee_models):
            model = getattr(sqlee_models, model)
            if hasattr(model, "repo"):
                print("正在读取Model '%s'" % (model.__class__.__name__))
                print("\n正在创建数据表: '%s'\n" % (model.tablename))
                repo.objects.create(name=model.tablename, namespace=model.namespaces)
    return True

class SYNC_DB(Thread):
    def __init__(self, repo=None, data=None, tablename=None):
        Thread.__init__(self)
        self.repo = repo
        self.data = data
        self.tablename = tablename

    def run(self):
        self.repo.objects.get(name=self.tablename).objects.create(**self.data)

def format_sql_data(sql_data):
    for data in sql_data:
        sql_data[data] = str(sql_data[data])
    return sql_data

def sync():
    from django.db import models
    repo = sqlee.connect(token=settings.SQLEE_SETTINGS["TOKEN"], repo=settings.SQLEE_SETTINGS["NAME"], owner=settings.SQLEE_SETTINGS["OWNER"])
    sys.path.append(settings.BASE_DIR)

    for app in settings.INSTALLED_APPS:
        try:
            print("正在解析模块'%s'内的DJANGO数据库声明文件." % (app))
            django_models = __import__("%s.models" % (app)).models
            sqlee_models = __import__("%s.sqlee_models" % (app)).sqlee_models
        except ModuleNotFoundError:
            print("模块'%s'中未找到DJANGO数据库声明文件." % (app))
            continue
        except AttributeError:
            print("模块'%s'中未找到DJANGO数据库声明文件." % (app))
            continue

        datas = []
        for model in dir(django_models):
            if not hasattr(eval("sqlee_models"), model) and not hasattr(eval("django_models"), model):
                continue
            sqlee_model = getattr(eval("sqlee_models"), model)
            model = getattr(eval("django_models"), model)

            if hasattr(model, "__base__"):
                print("正在读取Model '%s'" % (model.__name__))
                if model.__base__ is models.Model:
                    print("\n正在读取数据表: '%s'\n" % (sqlee_model().tablename))
                    sql_datas = model.objects.all()
                    for data in sql_datas:
                        datas.append(format_sql_data(model_to_dict(data)))
                        ### USES THREAD ###
                        #print(format_sql_data(model_to_dict(data)))
                        #sync = SYNC_DB(repo=repo, tablename=sqlee_model().tablename, data=format_sql_data(model_to_dict(data)))
                        #sync.setDaemon(True)
                        #sync.start()
                        ### USES THREAD ###
                        repo.objects.get(name=sqlee_model().tablename).objects.create(**format_sql_data(model_to_dict(data)))
                    #pool = SYNC_DB_POOL(repo=repo, datas=format_sql_data(model_to_dict(sql_datas)), tablename=sqlee_model().tablename)
                    #pool.setDaemon(True)
    return datas

def execute_from_command_line(argv):
    return eval(argv[1])()

verify_django()


#coding: utf-8
if __name__ == "__main__":
    from gitee import GiteeRepo
    from urlparse import URL
    from crypto import CryptoHandler
    from exceptions import (ColumnDataDoesNotFit, ColumnNotFoundError, 
                           DataWasNotGiven, DataWasNotGivenError)
else:
    from .gitee import GiteeRepo
    from .urlparse import URL
    from .crypto import CryptoHandler
    from .exceptions import (ColumnDataDoesNotFit, ColumnNotFoundError,
                            DataWasNotGiven, DataWasNotGivenError)

import json, warnings

def model_to_data(model=None):
    return tuple([i.data if hasattr(i, "data") else i for i in model])

class table_objects:
    def __init__(self, obj=None):
        self.obj = obj

    def all(self):
        """ Usage: all() """
        if not self.obj._loaded: self.obj.sync()
        return self.obj.columns
    
    def get(self, *args, **kwargs):
        """ Usage: get(**namespace) """
        if len(kwargs) == 1 and self.obj.index in kwargs:
            return SqleeColumn(
                    path = URL(self.obj.name) / kwargs[self.obj.index],
                    repo = self.obj.repo,
                    index = self.obj.index,
                    table = self.obj,
                    index_data = kwargs[self.obj.index],
                    namespace = self.obj.namespace,
                    )
        if not self.obj._loaded: self.obj.sync()
        need = len(kwargs)
        result = []
        for column in self.obj.columns:
            transit = 0
            for kwarg in kwargs:
                if kwargs[kwarg] in column.data:
                    transit += 1
            if transit == need:
                result.append(column)
        if len(result) != 1:
            raise ValueError("找到了 %d 个匹配的数据，如果你试图筛选多数据，请参考‘filter’." % len(result))
        return result[0]

    def create(self, *args, **kwargs):
        """ Usage: create(**namespace) """
        if self.obj.index == "id":
            if "id" in kwargs:
                raise ValueError("索引为 id 的表不应当将 id 作为数据参数传入.")
        else:
            if not self.obj.index in kwargs:
                raise DataWasNotGivenError(namespace=self.obj.index)

        column = self._get_cloumn(kwargs)

        all_columns = self.obj.repo.list_file(self.obj.name)
        if str(column) in all_columns:
            raise ValueError("目标索引已存在.")

        datas = []
        for namespace in self.obj.namespace:
            if namespace == self.obj.index:
                continue
            if namespace in kwargs:
                datas.append(kwargs[namespace])
            else:
                print(DataWasNotGiven(namespace=namespace))
                datas.append(None)

        self.obj.repo.upload_file(
            path = URL(self.obj.name) / column,
            content = self.obj.crypto.encrypt_to_str(json.dumps(datas))
            )

        self.obj.columns = list(self.obj.columns)
        self.obj.columns.append(
                SqleeColumn(
                    path = URL(self.obj.name) / column,
                    repo = self.obj.repo,
                    index = self.obj.index,
                    index_data = column,
                    table = self.obj,
                    namespace = self.obj.namespace,
                    )
                )
        self.obj.columns = tuple(self.obj.columns)
        return self.obj.columns

    def delete(self, *args, **kwargs):
        """ Usage: delete(**namespace) """
        force_del = kwargs["force_del"] if "force_del" in kwargs else False
        if "*" in args:
            if force_del:
                return self.obj.delete_all()
            else:
                raise ValueError("您选择了清空Table，然而却没有下定决心，略过.")
        if len(kwargs) == 1 and self.obj.index in kwargs:
            return self.obj.repo.delete_file(path = URL(self.obj.name) / kwargs[self.obj.index])
        if not self.obj._loaded: self.obj.sync()
        need = len(kwargs)
        result = []
        for column in self.obj.columns:
            transit = 0
            for kwarg in kwargs:
                if kwargs[kwarg] in column.data:
                    transit += 1
            if transit == need:
                result.append(column)
        if len(result) != 1 and not force_del:
            raise ValueError("找到了 %d 个匹配的数据，取消操作. 如果尝试删除多个数据，请加入参数force_del = True." % len(result))
        else:
            for column in result:
                self.obj.repo.delete_file(path = URL(self.obj.name) / column.index_data)
        return result[0]

        return self.obj.delete()
    
    @property
    def length(self):
        """ Usage: length """
        return len(self.obj.columns)

    def count(self):
        """ Usage: count() """
        return len(self.obj.columns)
    
    def filter(self, *args, **kwargs):
        """ Usage: filter(**namespace) """
        if len(kwargs) == 1 and self.obj.index in kwargs:
            if not kwargs[self.obj.index] in self.obj.repo.list_file(path=self.obj.name):
                print(ColumnNotFoundError(URL(self.obj.name) / kwargs[self.obj.index]))
                return []
            return [SqleeColumn(
                    path = URL(self.obj.name) / kwargs[self.obj.index],
                    repo = self.obj.repo,
                    index = self.obj.index,
                    table = self.obj,
                    index_data = kwargs[self.obj.index],
                    namespace = self.obj.namespace,
                    )]
        if not self.obj._loaded: self.obj.sync()
        need = len(kwargs)
        result = []
        for column in self.obj.columns:
            transit = 0
            for kwarg in kwargs:
                if kwargs[kwarg] in column.data:
                    transit += 1
            if transit == need:
                result.append(column)
        return result

    def _get_cloumn(self, kwargs):
        if self.obj.index == "id":
            column = self.obj.repo.list_file_int(path=self.obj.name)
            if len(column) == 0:
                column = 0
            else:
                column = max(column) + 1
            return column
        else:
            if self.obj.index in kwargs:
                return kwargs[self.obj.index]
            else:
                raise ValueError("未传入索引数据 %s." % self.obj.index)

    def sync(self):
        """ Usage: sync() """
        return self.obj.sync()

class SqleeColumn:
    datas = []
    def __init__(self, path=None, repo=None, index=None, index_data=None, 
                 table=None, namespace=[], *args, **kwargs):
        if not isinstance(repo, GiteeRepo):
            raise ValueError("参数'repo'必须是GiteeRepo.")

        self.index = index
        self.index_data = index_data
        self.repo = repo
        self.table = table
        self.path = path
        self.namespace = namespace
        self.sync()

    @property
    def data(self):
        """ Usage: data """
        return tuple(model_to_data(self.datas))

    @property
    def length(self):
        """ Usage: length """
        return len(self.data)

    def count(self):
        """ Usage: count() """
        return len(self.data)

    def sync(self):
        """ Usage: sync() """
        self.datas = [self.index_data, ]

        self.datas += json.loads(self.table.crypto.decrypt_by_str(
            self.repo.get_file(path=self.path)
            ))

        i = 0
        for data in self.datas:
            setattr(self, self.namespace[i], data)
            i += 1

        if len(self.datas) != len(self.namespace):
            print(ColumnDataDoesNotFit(namespace=self.namespace, datas=self.datas))
        return self.datas

    def update(self, *args, **kwargs):
        """ Usage: update(**namespace) """
        for kwarg in kwargs:
            self.datas[self.namespace.index(kwarg)] = kwargs[kwarg]
        self.repo.update_file(
            path = self.path,
            content = self.table.crypto.encrypt_to_str(json.dumps(self.datas[1:]))
            )
        return self.sync()

    def delete(self):
        """ Usage: delete() """
        self.repo.delete_file(path = self.path)
        del self
        return True

class SqleeTable:
    columns = []
    _loaded = False
    def __init__(self, name=None, repo=None, directly_load=True, encrypt=True):
        if not isinstance(name, str) and not isinstance(name, URL):
            raise ValueError("参数 'name' 必须是字符串或URL.")
        if not isinstance(repo, GiteeRepo):
            raise ValueError("参数 'repo' 必须是GiteeRepo.")

        self.crypto = CryptoHandler(crypto=encrypt)
        self.name = name
        self.repo = repo
        self.namespace = json.loads(self.repo.get_file(path=URL(self.name)/".namespace"))
        self.index = self.namespace[0]
        self.url = URL()/self.repo.user/self.repo.repo/self.name
        self.objects = table_objects(obj=self)
        if directly_load:
            self.sync()
    
    def get_column(self, id=None):
        for column in self.columns:
            if column.id == id:
                return column
        else:
            raise ColumnNotFoundError(id)

    def insert(self, *args, **kwargs):
        return self.objects.create(*args, **kwargs)

    def sync(self):
        self.columns = []
        all_columns = self.repo.list_file_int(path=self.name) if self.index == "id" else self.repo.list_file(path=self.name)
        for column in all_columns:
            self.columns.append(
                SqleeColumn(
                    path = URL(self.name) / column,
                    repo = self.repo,
                    index = self.index,
                    index_data = column,
                    table = self,
                    namespace = self.namespace,
                    )
                )
        self.columns = tuple(self.columns)
        self._loaded = True
        return self.columns

    def delete_all(self):
        answer = self.repo.drop_folder(path=self.name)
        del self
        return answer

if __name__ == "__main__":
    pass
    #repo = GiteeRepo(token="1895956f770eb0e4d08013ee4b753203", user="fu050409", repo="TEST_API")
    #table = SqleeTable(
    #    name="Table",
    #    repo = repo
    #    )
    #print(table.columns[0].datas[0].data)


import setuptools
from sqlee import config

__version__ = config.__version__
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name = "sqlee",
    version = __version__,
    author = "Entropy <fu050409@163.com>",
    author_email = "fu050409@163.com",
    description = "基于Gitee API的数据库.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://gitee.com/qu-c/sqlee",
    project_urls = {
        "Bug Tracker": "https://gitee.com/qu-c/sqlee/issues",
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license = "MIT",
    packages = setuptools.find_packages(),
    install_requires = [
        'requests',
        'prompt_toolkit',
    ],
    python_requires=">=3",
    
)



from Crypto.Cipher import AES
import base64
import binascii
import quopri
import json

class CryptoHandler:
    def __init__(self, key=None, iv=None, crypto=True):
        self.crypto = crypto
        if not self.crypto:
            return
        if key and not iv:
            self.key = key
            self.iv = key
        elif not key and iv:
            self.key = iv
            self.iv = iv
        elif key and iv:
            self.key = iv
            self.iv = iv
        else:
            self.key = b'.\x9c\xd0\xc31\xb0\x89\xe8\xc3\xf5\xc3\x89\xfcG\x1a6'
            self.iv = self.key

    def encrypt(self, string):
        if not self.crypto:
            return string
        cipher = AES.new(self.key, AES.MODE_CFB, self.iv)
        return cipher.encrypt(base64.b64encode(binascii.b2a_hex(quopri.encodestring(binascii.b2a_qp(string.encode())))))

    def decrypt(self, string):
        if not self.crypto:
            return string
        cipher = AES.new(self.key, AES.MODE_CFB, self.iv)
        return binascii.a2b_qp(quopri.decodestring(binascii.a2b_hex(base64.b64decode(cipher.decrypt(binascii.a2b_hex(string)))))).decode()

    def encrypt_to_str(self, string):
        if not self.crypto:
            return string
        cipher = AES.new(self.key, AES.MODE_CFB, self.iv)
        return binascii.b2a_hex(cipher.encrypt(base64.b64encode(binascii.b2a_hex(quopri.encodestring(binascii.b2a_qp(string.encode())))))).decode()

    def decrypt_by_str(self, string):
        if not self.crypto:
            return string
        cipher = AES.new(self.key, AES.MODE_CFB, self.iv)
        return binascii.a2b_qp(quopri.decodestring(binascii.a2b_hex(base64.b64decode(cipher.decrypt(binascii.a2b_hex(string.encode())))))).decode()

if __name__ == "__main__":
    crypto = CryptoHandler()
    print(crypto.encrypt_to_str(json.dumps({"key": "data"})))